# This TCL script is used by the freeWrap Makefile
#
# Purpose:
#     1) Create configuration file to be used at run-time by wrapped application.
#     2) Generate fwcrypt.c file
#     3) Add the standard TCL library directories to the tcl_pkgPath variable
#     4) Generate pkgIndex.tcl files so that TCL modules can be loaded with [package require]
#
#  Modified: October 6, 2019
#
if {$argv == {}} {
    set hdrFile "../src/fwcrypt.c"
    if {[catch {open $hdrFile w} fout]} {
        puts $fout
		set curpwd {}
       } {
           set charset {*+-.0123456789@ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_}
           set maxpos [string length $charset]
           set keylen [expr 8 + int(8 * rand())]
           set curpwd {}
           puts $fout "/* Automatically generated by setinfo.tcl - [clock format [clock seconds]]"
           puts $fout {}
           puts $fout "   This file defines the function that returns the encryption password."
           puts $fout "   Its contents have been randomly generated to produce a password"
           puts $fout "   that is difficult to extract from the compiled binary file."
           puts $fout "*/"
           puts $fout {}
           puts $fout {char *getPwdKey(char *keybuf)}
           puts $fout "\{"
           for {set idx 0} {$idx < $keylen} {incr idx} {
                set cval [string index $charset [expr int($maxpos * rand())]]
                scan $cval %c ival
                puts $fout "  keybuf\[$idx\] = $ival;"
                append curpwd $cval
               }
           puts $fout "  keybuf\[$keylen\] = 0;"
           puts $fout {}
           puts $fout "  return keybuf;"
           puts $fout "\}"
           close $fout
         }

    set pwdFile "../src/currentPWD.txt"
    if {[catch {open $pwdFile w} fout]} {
        puts $fout
       } {
           puts $fout $curpwd
           close $fout
         }
   } {
       ## Create the freeWrap initialization data file.
       set dirpath [lindex $argv 0]
       set progName [lindex $argv 1]
       set pLevel [lindex $argv 2]
       set configFile [file join $dirpath _freewrap_init.txt]
       if {[catch {open $configFile w} fout]} {
           _freewrap_message {.} warning ok {freeWrap error!} "Error creating $configFile.\n\n$fout."
           return 1
          }
       puts $fout /zvfs/freewrap.tcl
       puts $fout "$progName $pLevel"
       close $fout
       
       ## Modify init.tcl to properly initialize file path variables.
       set autolist {}
       set autolistTM {}
       set dirlist zipdir
       set laterlist {}
       set mountpt /zvfs

	   # Get list of subdirectories copied
	   set baselen [string length $dirlist]
	   while {[llength $dirlist]} {
			 set newdir [lindex $dirlist 0]
			 set newlist {}
			 if {[string length [glob -nocomplain $newdir/*.tm]]} {
			     # The directory contains a TCL module.
			     puts "Processing TCL module directory $newdir"
			     lappend autolistTM [string range $newdir $baselen end]
			    } {
				  puts "Processing TCL package directory $newdir"
				  set pkgFile [file join $newdir pkgIndex.tcl]
				  if {[file exists $pkgFile]} {
				      set addPath [string range $newdir $baselen end]
					  if {[string range $addPath 0 4] eq {/tcl8} || [string range $addPath 0 3] eq {/tk8} || [string range $addPath 0 5] eq {/twapi} || [string range $addPath 0 6] eq {/winico}} {
						  lappend autolist $mountpt$addPath
						 } { lappend laterlist $mountpt$addPath }
				    }
				  set newlist [glob -nocomplain -types d [file join $newdir *]]
			      }
			 if {$newlist eq {}} {
				set dirlist [lrange $dirlist 1 end]
			    } { set dirlist [concat [lrange $dirlist 1 end] $newlist] }
			}

	   # Add list of subdirectories to tcl_pkgPath in init.tcl file.
       puts "Modifying init.tcl with new module paths."
	   set initFile [file join zipdir tcl8.6 init.tcl]
	   
	   set fin [open $initFile r]
	   set data [read $fin]
	   close $fin

	   set fout [open $initFile w]
	   set FWsignature {## FreeWrap modifications follow}
       set origLen [string first $FWsignature $data]
       if {$origLen == -1} {
           puts -nonewline $fout $data
          } {
              incr origLen -1
              puts -nonewline $fout [string range $data 0 $origLen]
            }
		  puts $fout $FWsignature
          puts $fout {# Set up TCL auto_path variable.}
          puts $fout {global auto_path}
		  foreach dir $autolist {
			  set newpath [file join $mountpt $dir]
			  puts $fout "lappend ::auto_path \{$newpath\}"
			}
		  puts $fout {}

		  # Extend auto_path to find standard TCL packages
		  puts $fout {# Point to standard TCL packages}
		  puts $fout {set ::tcl_pkgPath {}}
		  foreach dir $autolist {
			  set newpath [file join $mountpt $dir]
			  puts $fout "lappend ::tcl_pkgPath \{$newpath\}"
			}
		  puts $fout {}

		  # Set up TCL module load paths.
		  puts $fout {# Point to wrapped tcl modules}
		  puts $fout {foreach dtemp [::tcl::tm::path list] {::tcl::tm::path remove $dtemp}}
		  foreach dir $autolistTM {
			  set newpath $mountpt$dir
			  puts $fout "::tcl::tm::path add \{$newpath\}"
			}
		  puts $fout {}
        
          # Create the get_laterlist procedure.
		  puts $fout {# Create procedure that can be used to extend path variables later.}
		  puts $fout "proc ::get_laterlist \{\} \{"
		  puts $fout "return \{$laterlist\}\n\}"
		  close $fout
     }

exit

